import logging
import traceback
from time import perf_counter
from typing import List, Optional

import pandas as pd  # type: ignore

from mate.alerts import Alert, AlertTarget, FeatureAlertKind, InferenceException
from mate.db import (
    Feature,
    FeatureAlert,
    NumericalStats,
    StringStats,
    get_current_mate,
    get_features,
    get_statistics,
)
from mate.stats import FeatureType, Statistics

MATE_STATISTICS_PATH_VAR = "MATE_STATISTICS_PATH"
MATE_CONSTRAINTS_PATH_VAR = "MATE_CONSTRAINTS_PATH"
OUTLIER_CUTOFF = 4.0


logger = logging.getLogger("mate")


class RunningMate(object):
    mate_name: str
    statistics: Optional[Statistics]
    feature_alerts: List[FeatureAlert]
    targets: List[AlertTarget]

    def __init__(
        self,
        mate_name: str,
        mate_version: int,
        df: pd.DataFrame,
        targets: List[AlertTarget],
    ):
        self.mate_name = mate_name
        self.mate_version = mate_version
        self.targets = targets

        self.mate = get_current_mate(self.mate_name)

        if self.mate:
            self.features = get_features(self.mate)

        logger.info(f"\nUsing model version {self.mate_version}.")

        self.feature_alerts: List[FeatureAlert] = []
        self.feature_alerts.extend(self._check_statistics(df))

    def _check_statistics(self, df: pd.DataFrame) -> List[FeatureAlert]:
        result = []

        for feature in self.features:
            col = df[feature.name]
            result.extend(self._check_feature_statistics(feature, col))

        logger.info(f"Found {len(result)} statistical alerts.")

        return result

    def _check_feature_statistics(
        self, feature: Feature, col: pd.Series
    ) -> List[FeatureAlert]:
        result: List[FeatureAlert] = []
        num_missing = None

        if feature.inferred_type == FeatureType.INTEGRAL.value:
            numerical_statistic = get_statistics(NumericalStats, feature)

            # outlier check
            if (
                (abs(col - numerical_statistic.mean) / numerical_statistic.std_dev)
                > OUTLIER_CUTOFF
            ).any():
                result.append(
                    FeatureAlert.create(
                        name=col.name,
                        kind=FeatureAlertKind.OUTLIER.value,
                        value=col.item(),
                        feature=feature,
                    )
                )

            # bound check
            if (col < numerical_statistic.min).any() or (
                col > numerical_statistic.max
            ).any():
                result.append(
                    FeatureAlert.create(
                        name=col.name,
                        kind=FeatureAlertKind.BOUND.value,
                        value=col.item(),
                        feature=feature,
                    )
                )

            num_missing = numerical_statistic.num_missing

        if feature.inferred_type == FeatureType.STRING.value:
            string_statistic = get_statistics(StringStats, feature)

            num_missing = string_statistic.num_missing

        # null check
        if num_missing is not None and num_missing == 0:
            if col.isnull().any():
                result.append(
                    FeatureAlert.create(
                        name=col.name,
                        kind=FeatureAlertKind.NULL.value,
                        value=col.item(),
                        feature=feature,
                    )
                )

        return result

    def __enter__(self):
        self.time_start = perf_counter()

    def __exit__(self, type, value, tb):
        time_end = perf_counter()
        logger.info(f"Elapsed inference time in seconds: {time_end - self.time_start}")

        inference_exception = None

        if type is not None:
            logger.exception(f"Exception occurred during model inference: {value}")
            inference_exception = InferenceException(
                message=value,
                traceback="\n".join(traceback.format_exception(type, value, tb)),
            )

        alert = Alert(self.mate_name, self.feature_alerts, inference_exception)

        if len(alert.features) > 0 or alert.exception is not None:
            for target in self.targets:
                target.send_alert(alert)



import pandas as pd  # type: ignore

from mate.db import Feature, NumericalStats, StringStats, get_current_mate
from mate.stats import (
    CommonStatistics,
    FeatureStatistics,
    FeatureType,
    NumericalStatistics,
    Statistics,
    StringStatistics,
)


def generate_baseline_stats(df: pd.DataFrame, name: str, version: int):
    statistics = _gen_statistics(df)

    mate = get_current_mate(name)
    if mate:
        mate.item_count = statistics.item_count
        mate.save()

    for feature in statistics.features:
        feature_stats = Feature(
            name=feature.name, inferred_type=feature.inferred_type, mate=mate
        )
        feature_stats.save()

        if feature.numerical_statistics:
            NumericalStats.create(
                num_present=feature.numerical_statistics.common.num_present,
                num_missing=feature.numerical_statistics.common.num_missing,
                mean=feature.numerical_statistics.mean,
                sum=feature.numerical_statistics.sum,
                std_dev=feature.numerical_statistics.std_dev,
                min=feature.numerical_statistics.min,
                max=feature.numerical_statistics.max,
                feature=feature_stats,
            )

        if feature.string_statistics:
            StringStats.create(
                num_present=feature.string_statistics.common.num_present,
                num_missing=feature.string_statistics.common.num_missing,
                distinct_count=feature.string_statistics.distinct_count,
                feature=feature_stats,
            )


def _create_statistics_feature(feature_series: pd.Series) -> FeatureStatistics:
    feature_name = feature_series.name
    feature_type = _infer_feature_type(feature_series)

    feature = FeatureStatistics(name=feature_name, inferred_type=feature_type.value)

    n_missing = feature_series.isna().sum()
    n_present = len(feature_series) - n_missing

    common = CommonStatistics(n_present, n_missing)

    if feature_type in {FeatureType.INTEGRAL, FeatureType.FRACTIONAL}:
        feature.numerical_statistics = NumericalStatistics(
            common=common,
            mean=feature_series.mean(),
            sum=feature_series.sum(),
            std_dev=feature_series.std(),
            min=feature_series.min(),
            max=feature_series.max(),
        )

    elif feature_type == FeatureType.STRING:
        feature.string_statistics = StringStatistics(
            common=common, distinct_count=len(feature_series.dropna().unique())
        )

    return feature


def _gen_statistics(df: pd.DataFrame) -> Statistics:
    statistics = Statistics(
        item_count=len(df),
        features=[
            _create_statistics_feature(feature_series)
            for name, feature_series in df.iteritems()
        ],
    )
    return statistics


def _infer_feature_type(feature_series: pd.Series) -> FeatureType:
    dtype_name = str(feature_series.dtype)

    # {"int8", "int16", "int32", "int64", "intp"}
    # {"uint8", "uint16", "uint32", "uint64", "uintp"}
    if dtype_name.startswith("int") or dtype_name.startswith("uint"):
        feature_type = FeatureType.INTEGRAL

    # {"float16", "float32", "float64", "float96", "float128"}:
    elif dtype_name.startswith("float"):
        feature_type = FeatureType.FRACTIONAL

    # {"string", "<U16/32/...", ">U16/32/...", "=U16/32/..."}
    elif (dtype_name == "string") or (dtype_name[:2] in {"<U", ">U", "=U"}):
        feature_type = FeatureType.STRING

    elif dtype_name == "object":
        feature_type = FeatureType.UNKNOWN

        # attempt to infer if object dtype is actually a string
        types = set(map(type, feature_series.dropna()))
        if types == {str}:
            feature_type = FeatureType.STRING

    else:
        # Bools, datetimes, etc are all treated as unknown
        feature_type = FeatureType.UNKNOWN

    return feature_type
